'******************************************************************************
'*  io.spin - handle input and output requests sent by pm80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.6 2010-06-05 05:59:21 pm Exp $
'******************************************************************************
CON

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read keyboard

	baud			=	115200		' baud rate
	SCRATCH_RAM		=	$4000
	FSEL_TOP		=	2
	FSEL_H			=	23

VAR
	long	cog
	long	rc
	long	params[8]
	word	topy
	word	boty
	word	ndir
	word	row
	word	y0
	word	speed
	byte	curx
	byte	cury
	byte	getdir
	byte	de[32]					' directory entry
	byte	name[12+1]
	byte	buff[64]

OBJ
	fat	: "fatfs"
	vga	: "vga"
	kbd	: "keyboard"
	ay	: "AYcog_V0-22"

PUB start(pins) : okay | i, j, k, need_rom
	stop

	iocmd_ptr := @io_command
	videoram_ptr := @videoram
	startaddr_ptr := @startaddr
	crtc_ptr := @crtc
	psg_ptr := @psg
	portff_ptr := @portff
	sync_ptr := @sync

	' ***************************************************************
	' Start the FAT driver and the SD SPI object (must be first)
	'
	okay := \fat.start(byte[pins][2], byte[pins][3], byte[pins][4], byte[pins][5])
	if okay < 0
		failure(string("FAT + SD SPI"),okay)
		return -1
	iobuff_ptr := \fat.get_memory
	spi_command := fat.get_SPI_command
	spi_blocknr := spi_command + 4
	spi_address := spi_command + 8

	' ***************************************************************
	' Start the Colour Genie VGA video object
	' pass screen ptr, colour ptr, font ptr, cursor ptr, sync ptr
	'

	params[0] := @videoram
	params[1] := @startaddr
	params[2] := @colourram
	params[3] := @fontram
	params[4] := @crtc
	params[5] := @psg
	params[6] := @portff
	params[7] := @sync
	vga.start(byte[pins][9], @params)

	chr(12)
	str(string("Colour Genie EG2000 emulator"))
	crlf

	' ***************************************************************
	' Start the Colour Genie keyboard driver
	'
	okay := \kbd.startx(byte[pins][6], byte[pins][7], 0, 40)
	if okay < 0
		failure(string("KBD"),okay)
		return
	kbd_states := kbd.states_ptr

	' ***************************************************************
	' Start the AY-3-8910 emulator cog
	'
	okay := \ay.start(byte[pins][8], 0, psg_ptr)
	if okay < 0
		failure(string("PSG"),okay)
		return
	byte[psg_ptr][7] := $3f

	' ***************************************************************
	' Start the I/O handler cog
	'
	okay := cog := COGNEW(@entry, 0) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1

	' ***************************************************************
	' mount the FAT, i.e. check the format of the SD media
	'
	okay := \fat.mount(0)
	if okay < 0
		failure(string("FAT"), okay)
		return -1

	rom_load

	' *********************************************************
	' file selection
	'
	waitcnt(CNT + 2*clkfreq)

	getdir := 1
	repeat y0 from FSEL_TOP - 1 to 15
		gotoxy(0, y0)
		chr(22)
	repeat while getdir
		gotoxy(0, FSEL_TOP - 1)
		str(string("Reading directory...",22))
		readdir
		row := 0
		topy := 0
		gotoxy(0, FSEL_TOP - 1)
		str(string("Select a file to load ("))
		dec(boty)
		chr("/")
		dec(ndir)
		str(string("):"))
		repeat
			repeat y0 from 0 to FSEL_H - 1
				rc := direntry_name(topy + y0, @buff)
				if rc < 0
					buff[0] := " "
					buff[1] := " "
					buff[2] := 0
				gotoxy(0, FSEL_TOP+y0)
				if y0 == row
					str(string("=> "))
				else
					str(string("   "))
				str(@buff)
				if y0 == row
					str(string(" <="))
				chr(22)

			waitcnt(CNT + clkfreq >> 4)

			if kbd.keystate($34)	' Down arrow
				if (row < FSEL_H - 1) and (row < boty)
					row++
				elseif topy < boty - (FSEL_H - 1)
					topy++

			if kbd.keystate($33)	' Up arrow
				if row > 0
					row--
				elseif topy > 0
					topy--

			if kbd.keystate($3b)	' Page down
				if topy + FSEL_H + FSEL_H - 1 < boty
					topy += FSEL_H
				elseif boty => FSEL_H
					topy := boty - (FSEL_H - 1)
					row := boty - topy
				else
					topy := 0
					row := boty

			if kbd.keystate($3a)	' Page up
				if topy - FSEL_H > 0
					topy -= FSEL_H
				else
					topy := 0
					row := 0

			if kbd.keystate($30)	' Return
				repeat y0 from 0 to FSEL_H - 1
					gotoxy(0, FSEL_TOP+y0)
					chr(22)
				gotoxy(0, FSEL_TOP)
				getdir := 0
				rc := direntry_name(topy + row, @buff)
				if byte[@buff][0] == "<"
					chdir_entry(topy + row)
					getdir := 1
				quit
		
	' *********************************************************
	' load the selected file
	'
	cmd_load(@buff)
	waitcnt(cnt + clkfreq)

	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PUB get_iocmd_ptr
	return @io_command

PUB get_irq
	return @irq

PUB get_videoram
	return @videoram

PUB get_startaddr
	return @startaddr

PUB get_colourram
	return @colourram

PUB get_fontram
	return @fontram

PUB get_crtc
	return @crtc

PUB get_psg
	return @psg

PUB get_portff
	return @portff

PUB get_sync
	return @sync

PRI gotoxy(x,y)
	curx := (x #> 0) <# 39
	cury := (y #> 0) <# 24

PRI chr(ch)
	case ch
		22:	' clear to end of line
			bytefill(videoram_ptr + cury * 40 + curx, $20, 39 - curx)
		13:	' carriage return
			curx := 1
		12:	' form feed - clear screen
			longfill(videoram_ptr, $20202020, 40*24/4)
			curx := 0
			cury := 0
		10:	' line feed
			cury += 1
			if cury > 24
				longmove(videoram_ptr, videoram_ptr + 40, 40*24/4)
				longfill(videoram_ptr + 40*24, $20202020, 40/4)
				cury := 15
		9:	' tab
			repeat until (curx & 7) == 0
				chr(32)
		8:	' back space
			if curx > 0
				--curx
		other:
			byte[videoram_ptr][cury * 40 + curx] := ch
			curx++
			if curx == 40
				chr(13)
				chr(10)

PRI str(src) | i
	repeat i from 0 to strsize(src) - 1
		chr(byte[src][i])

PRI dec(value) | _i
'' Print a decimal number
	result := 0
	if value < 0
		chr("-")
		if (value == NEGX)
			chr("2")
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PRI crlf
	chr(13)
	chr(10)

PRI dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PRI hex(value, digits)
'' Print a hexadecimal number

	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PRI bin(value, digits)
'' Print a binary number

	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI chr_ptr(ch, dst)
	byte[dst++] := ch
	byte[dst] := 0
	return dst

PRI str_ptr(src, dst) | i
	repeat i from 0 to strsize(src) - 1
		byte[dst++] := byte[src][i]
	byte[dst] := 0
	return dst

PRI dec_ptr(value, dst) | _i
'' Print a decimal number
	result := 0
	if value < 0
		dst := chr_ptr("-", dst)
		if (value == NEGX)
			dst := chr_ptr("2", dst)
			value += 2_000_000_000
		-value
	_i := 1_000_000_000
	repeat 10
		if value => _i
			dst := chr_ptr(value / _i + "0", dst)
			value //= _i
			result~~
		elseif result or _i == 1
			dst := chr_ptr("0", dst)
		_i /= 10
	return dst

PRI dec_size_ptr(value, dst)
	if value > 1048576
		dst := dec_ptr(value/1048576, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1048576) / 1048576, dst)
		dst := str_ptr(string("MB"), dst)
	elseif value > 1024
		dst := dec_ptr(value/1024, dst)
		dst := chr_ptr(".", dst)
		dst := dec_ptr(10*(value // 1024) / 1024, dst)
		dst := str_ptr(string("KB"), dst)
	else
		dst := dec_ptr(value, dst)
	return dst

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who)
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI strncmp(dst, src, n) | i
	repeat i from 0 to n - 1
		if byte[dst][i] <> byte[src][i]
			return byte[dst][i] - byte[src][i]
	return 0

PRI rom_load | fp, i
	' *********************************************************
	' find the CGENIE.ROM file and load it into the ROM region
	'
	fp := \fat.fopen(string("CGENIE.ROM"), "r")
	if fp < 0
		return -1
	repeat i from $0000 to $3e00 step $200
		if fat.fread(fp, iobuff_ptr, $200) == $200
			fat.write_xmm_ram(iobuff_ptr, i, $200)
	fat.fclose(fp)
	return 0

PRI readdir | rec, i, swapped
	rec := fat.opendir(@de)
	gotoxy(0,0)
	boty := 0
	ndir := 0
	repeat while rec => 0
		ndir++
		if byte[@de][fat#dent_name] <> $e5	' deleted entry?
			if fat.match(@de, string("*.CMD")) or (fat.isdir(@de) and not fat.isdot(@de))
				fat.write_xmm_ram(@de, SCRATCH_RAM + boty * 32, 32)
				boty++
		rec := fat.readdir(rec, @de)
	' do a bubble sort of the filenames in memory
	swapped := 1
	repeat until swapped == 0
		swapped := 0
		repeat i from 0 to boty - 2
			' read two directory entries
			fat.read_xmm_ram(@buff, SCRATCH_RAM + i * fat#dent_bytes, 2*fat#dent_bytes)
			' and compare the names
			if strncmp(@buff, @buff + fat#dent_bytes, 8) > 0
				fat.write_xmm_ram(@buff, SCRATCH_RAM + (i + 1) * fat#dent_bytes, fat#dent_bytes)
				fat.write_xmm_ram(@buff + fat#dent_bytes, SCRATCH_RAM + i * fat#dent_bytes, fat#dent_bytes)
				swapped++

PRI direntry_name(n, buf) | i, temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * 32, 32)
		if (byte[@temp][fat#dent_attr] & fat#dent_attr_D) <> 0
			' directory
			byte[buf] := "<"
			fat.get_dent_name(@temp, buf + 1)
			i := strsize(buf)
			byte[buf + i + 0] := ">"
			byte[buf + i + 1] := 0
		else
			fat.get_dent_name(@temp, buf)
		return 0
	return -1

PRI chdir_entry(n) | temp[fat#dent_bytes/4]
	if n < boty
		fat.read_xmm_ram(@temp, SCRATCH_RAM + n * fat#dent_bytes, fat#dent_bytes)
		fat.chdir_dent(@temp)
		return 0
	return -1

PRI open_file(fname_ptr) : dent | size, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		str(fname_ptr)
		str(string(" not found.",13,10))
		return -1
	bytemove(@de, dent, 32)
	fat.get_dent_name(dent, @name)
	size := fat.get_dent_size(@de)
	sec := fat.get_dent_first_sector(@de)
	str(@name)
	str(string(9,"sector "))
	dec(sec)
	str(string(9,"size "))
	dec_size(size)
	crlf

PRI cmd_load(fname_ptr) | dent, sec, nsec, pos, state, badr, blen, bmin, bmax
	dent := open_file(fname_ptr)
	if dent < 0
		return -1
	fat.get_dent_name(dent, @name)
	sec := fat.get_dent_first_sector(@de)

	pos := $200
	state := 0
	bmin := $ffff
	bmax := $0000
	repeat
		if pos == $200
			if sec < 0
				quit
			nsec := fat.get_dent_next_sector(@de, sec)
			fat.read_sector(sec)
			sec := nsec
			pos := 0
		case state
			0:	' expecting a header
				case byte[iobuff_ptr][pos++]
					$01:	' CMD load block header
						state := 1
					$02:	' CMD entry point header
						state := 7
					other:	' CMD comment block
						state := 5

			1:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 2

			2:	' expecting address LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 3

			3:	' expecting address MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				state := 4
				if blen =< 0
					blen += 256
				bmin := bmin <# badr

			4:	' expecting blen bytes of data
				if pos + blen =< $200
					fat.write_xmm_ram(iobuff_ptr + pos, badr, blen)
					badr += blen
					pos += blen
					blen := 0
				else
					fat.write_xmm_ram(iobuff_ptr + pos, badr, $200 - pos)
					badr += $200 - pos
					blen -= $200 - pos
					pos := $200
				if blen == 0
					bmax := bmax #> badr
					state := 0

			5:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				if blen == 0
					blen := 256
				state := 6

			6:	' ignoring comment data
				if pos + blen <= $200
					pos += blen
					blen := 0
				else
					blen -= $200 - pos
					pos := $200
				if blen == 0
					state := 0

			7:	' expecting a block length
				blen := byte[iobuff_ptr][pos++]
				state := 8

			8:	' expecting entry point LSB
				blen--
				badr := byte[iobuff_ptr][pos++]
				state := 9

			9:	' expecting entry point MSB
				blen--
				badr += 256 * byte[iobuff_ptr][pos++]
				str(string("Loaded from "))
				hex(bmin, 4)
				str(string(".."))
				hex(bmax - 1, 4)
				crlf
				str(string("Entry point "))
				hex(badr, 4)
				crlf
				fat.write_xmm_ram(@badr, $0116, 2)
				quit

DAT
'---------------------------------------------------------------------------------------------------------

		org	0

'---------------------------------------------------------------------------------------------------------
entry
			jmp	#run

iocmd_ptr		long	0-0
iobuff_ptr		long	0-0
videoram_ptr		long	0-0
startaddr_ptr		long	0-0
screen_size		long	(40*108) << 16		' maximum screen size
crtc_ptr		long	0-0			' pointer to CRTC 6845 registers
psg_ptr			long	0-0			' pointer to AY-3-8910 registers
portff_ptr		long	0-0
sync_ptr		long	0-0
crtc_idx		long	0			' Index of CRTC register
psg_idx			long	0			' Index of PSG register
addr			long	0
cur_x			long	0
cur_y			long	0
x20202020		long	$20202020
pff			long	%00011000

spi_command		long	0-0
spi_blocknr		long	0-0
spi_address		long	0-0
	
kbd_states		long	0-0
kbd_index		long	0
kbd_shift		long	0
kbd_data		long	0

minus1			long	$ffffffff
low_24bits		long	$00ffffff
low_16bits		long	$0000ffff
msg_break_ptr		long	@@@msg_break

iocmd			long	0
port			long	0
data			long	0
null			long	0
vsync			long	0

t1			long	0
t2			long	0
t3			long	0
t4			long	0

char			long	0
strptr			long	0
hexval			long	0
hexcnt			long	0

run
			wrlong	null, iocmd_ptr

command
			rdlong	vsync, sync_ptr		wz
		if_nz	jmp	#do_sync
			rdlong	iocmd, iocmd_ptr
			mov	t1, iocmd
			shr	t1, #24			wz
		if_z	jmp	#command

			cmp	t1, #io_cmd_out		wz
		if_z	jmp	#output
			cmp	t1, #io_cmd_in		wz
		if_z	jmp	#input
			cmp	t1, #io_cmd_break	wz
		if_z	jmp	#break
			cmp	t1, #io_cmd_kbd		wz
		if_z	jmp	#key_matrix
command_0
			and	iocmd, low_24bits
			wrlong	iocmd, iocmd_ptr
			jmp	#command

'*******************************************
'
' Do vertical sync handling
'
do_sync
			mov	vsync, #0
			wrlong	vsync, sync_ptr			' clear vsync flag
			jmp	#command

'*******************************************
'
' Read i/o port and data to write to
'
output
			mov	port, iocmd			' get command word
			shr	port, #8			' extract port
			and	port, #$ff			' lower 8 bits
			mov	data, iocmd			' get command word
			and	data, #$ff			' extract data
  	
			cmp	port, #$f8    		wz	' PSG index?
		if_z	jmp	#output_psg_index
			cmp	port, #$f9		wz	' PSG data?
		if_z	jmp	#output_psg_data
			cmp	port, #$fa    		wz	' CRTC index?
		if_z	jmp	#output_crtc_index
			cmp	port, #$fb		wz	' CRTC data?
		if_z	jmp	#output_crtc_data
			cmp	port, #$ff		wz	' port = $FF?
		if_z	jmp	#output_ff

			jmp	#command_0

output_ff
			mov	pff, data			' save value
			wrlong	pff, portff_ptr
			' TODO: check the bits of data for cassette?
			jmp	#command_0

output_psg_index
			mov	psg_idx, data
			jmp	#command_0

output_psg_data
			cmp	psg_idx, #16		wz, wc	' writeable register?
		if_ae	jmp	#command_0
			mov	t1, psg_idx
			add	t1, psg_ptr
			cmp	psg_idx, #13		wz
		if_z	and	data, #15			' make sure we don't set the trigger bit
			wrbyte	data, t1
			jmp	#command_0

output_crtc_index
			mov	crtc_idx, data
			jmp	#command_0

output_crtc_data
			cmp	crtc_idx, #18		wz, wc	' index out of bounds?
		if_ae	jmp	#command_0			' yes, ignore the write
			mov	t1, crtc_idx
			add	t1, crtc_ptr
			rdbyte	t2, t1
			cmp	data, t2		wz
		if_z	jmp	#command_0			' no change?
			wrbyte	data, t1
			cmp	crtc_idx, #$0c		wz	' start address high?
		if_nz	cmp	crtc_idx, #$0d		wz	' start address low?
		if_nz	jmp	#command_0			' none of them

			mov	t1, #$0c
			add	t1, crtc_ptr
			rdword	t2, t1				' get start address high + low
			andn	t2, #$c0
			or	t2, #$40			' video memory is $4000-$7fff
			ror	t2, #8
			mov	t1, t2
			shr	t1, #16
			and	t2, #$ff
			or	t2, t1
			wrlong	t2, startaddr_ptr
			wrlong	t2, spi_blocknr			' set source address
			mov	t2, screen_size			' copy (40*108) bytes
			or	t2, videoram_ptr
			wrlong	t2, spi_address			' set hub RAM address + length
			mov	t2, #"R"			' XMM read command
			wrlong	t2, spi_command
:wait			rdlong	t2, spi_command		wz
		if_nz	jmp	#:wait
			jmp	#command_0

'*******************************************
'
' Read i/o port to read from
'
input
			andn	iocmd, #$ff			' clear the input data
			mov	port, iocmd			' get command word
			shr	port, #8
			and	port, #$ff
	
			cmp	port, #$f8		wz	' PSG index?
		if_z	jmp	#input_psg_data
			cmp	port, #$f9		wz	' PSG data?
		if_z	jmp	#input_psg_data
			cmp	port, #$fa    		wz	' CRTC index?
		if_z	jmp	#input_crtc_index
			cmp	port, #$fb		wz	' CRTC data?
		if_z	jmp	#input_crtc_data
			cmp	port, #$ff		wz	' port = $FF?
		if_z	jmp	#input_ff
	
			or	iocmd, #$ff			' return $ff for other ports
			jmp	#command_0

input_psg_index
			mov	data, psg_idx
input_data
			or	iocmd, data
			jmp	#command_0

input_psg_data
			mov	data, #$ff
			cmp	psg_idx, #16		wz, wc	' beyond readable registers?
		if_ae	jmp	#input_data
			mov	t1, psg_idx
			add	t1, psg_ptr
			rdbyte	data, t1
			jmp	#input_data

input_crtc_index
'			mov	data, crtc_idx
'			jmp	#input_data

input_crtc_data
			mov	data, #$ff
			cmp	crtc_idx, #18		wz, wc	' index out of bounds?
		if_ae	jmp	#input_data
			cmp	crtc_idx, #$0a		wz, wc	' write-only register?
		if_b	jmp	#input_data
			mov	t1, crtc_idx
			add	t1, crtc_ptr
			rdbyte	data, t1			' read the CRTC register
			jmp	#input_data

input_ff
			rdlong	pff, portff_ptr
			mov	data, pff
			jmp	#input_data
		
'*******************************************
'
' Output a character to the screen
'
conout_chr
			mov	addr, cur_y			' y
			shl	addr, #2			' * 4
			add	addr, cur_y			' * 5
			shl	addr, #3			' * 40
			add	addr, cur_x			' + x
			add	addr, videoram_ptr		' + screen address

			cmp	char, #8		wz	' backspace?
		if_z	jmp	#:bsp
			cmp	char, #9		wz	' tab?
		if_z	jmp	#:tab
			cmp	char, #10		wz	' line feed?
		if_z	jmp	#:lf
			cmp	char, #12		wz	' clear screen?
		if_z	jmp	#:cls
			cmp	char, #13		wz	' carriage return?
		if_z	jmp	#:cr
:out
			wrbyte	char, addr
			add	cur_x, #1
:wrap
			cmp	cur_x, #40		wz
		if_nz	jmp	#:exit
			mov	cur_x, #0
:lf
			add	cur_y, #1
			cmp	cur_y, #16		wz
		if_nz	jmp	#:exit
			call	#scroll
			mov	cur_y, #15
			jmp	#:exit

:bsp
			cmp	cur_x, #0		wz, wc
		if_be	jmp	#:exit
			sub	cur_x, #1
			jmp	#:exit
:tab
			mov	data, #32
			wrbyte	data, addr
			add	addr, #1
			add	cur_x, #1
			test	cur_x, #7		wz
		if_nz	jmp	#:tab
			jmp	#:wrap
:cr
			mov	cur_x, #0
			jmp	#:exit
:cls
			mov	addr, videoram_ptr
			mov	cur_y, #40*25/4
:loop
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_y, #:loop
			mov	cur_x, #0
:exit
conout_chr_ret
			ret		

'*******************************************
'
' Scroll the screen one line up
'
scroll
			mov	addr, videoram_ptr
			mov	cur_y, #24
:row
			mov	cur_x, #40/4
:col
			add	addr, #40
			rdlong	char, addr
			sub	addr, #40
			wrlong	char, addr
			add	addr, #4
			djnz	cur_x, #:col
			djnz	cur_y, #:row
			mov	cur_x, #40/4
:fill
			wrlong	x20202020, addr
			add	addr, #4
			djnz	cur_x, #:fill
scroll_ret
			ret


'*******************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
			mov	hexcnt, #2
			shl	hexval, #24
			jmp	#conout_hex

'*******************************************
'
' Output a word in hexval to the console
'
conout_hex_w
			mov	hexcnt, #4
			shl	hexval, #16
			jmp	#conout_hex

'*******************************************
'
' Output a long in hexval to the console
'
conout_hex_l
			mov	hexcnt, #8
conout_hex
			mov	char, hexval
			shr	char, #28
			cmp	char, #10	wz, wc
		if_b	add	char, #48
		if_ae	add	char, #55
			call	#conout_chr
			shl	hexval, #4
			djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
			ret

'*******************************************
'
' Output a carriage return to the console
'
conout_space
			mov	char, #32
			call	#conout_chr
conout_space_ret
			ret

'*******************************************
'
' Output a carriage return to the console
'
conout_crlf
			mov	char, #13
			call	#conout_chr
			mov	char, #10
			call	#conout_chr
conout_crlf_ret
			ret

'*******************************************
'
' Output a string at strptr to the console
'
conout_string
			rdbyte	char, strptr	wz
			add	strptr, #1
		if_z	jmp	#conout_string_ret
			call	#conout_chr
			jmp	#conout_string
conout_string_ret
			ret

'*******************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
			mov	t1, iocmd			' read the address of the register file
			shr	t1, #8				' bits 23..8 to 15..0
			and	t1, low_16bits			' mask only bits 15..0
			mov	strptr, msg_break_ptr	

			mov	t2, #8
:loop_1			call	#conout_string
			rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
			call	#conout_hex_w
			add	t1, #2
			djnz	t2, #:loop_1

			call	#conout_string			' new line and print "R:"
			rdbyte	hexval, t1			' R
			add	t1, #1
			rdbyte	t2, t1				' R2
			and	hexval, #%01111111		' R register's bit #7
			or	hexval, t2			' does never change from the last written value
			call	#conout_hex_b
			add	t1, #1
	
			call	#conout_string
			rdbyte	hexval, t1			' IFF1+2/IM
			mov	t2, t1				' save value
			and	hexval, #1			' IFF1 (interrupt flip-flop 1)
			call	#conout_hex_b
	
			call	#conout_string
			mov	hexval, t2			' IFF1+2/IM
			shr	hexval, #1			' shift bit #1 to #0
			and	hexval, #1			' IFF2 (interrupt flip-flop 2)
			call	#conout_hex_b
	
			call	#conout_string
			mov	hexval, t2			' IFF1+2/IM
			shr	hexval, #2			' shift bits #3..2 to #1..0
			and	hexval, #3			' IM (interrupt mode)
			call	#conout_hex_b
			add	t1, #1
	
			call	#conout_string
			rdbyte	hexval, t1			' I (interrupt vector register)
			call	#conout_hex_b
			add	t1, #1
	
			mov	t2, #4
:loop2			call	#conout_string
			rdword	hexval, t1			' IX
			mov	t3, hexval
			call	#conout_hex_w
			add	t1, #2
			djnz	t2, #:loop2
			call	#conout_crlf

			jmp	#command_0

'*******************************************
'
' Read the TRS-80 keyboard matrix
' Address is in io_port, address bits 0..7
' determine the keyboard row(s) to read
'
key_matrix
			mov	port, iocmd			' get command word
			shr	port, #8
			and	port, low_16bits
			mov	t1, #0				' resulting bits
			mov	t2, kbd_states			' address of the keyboard state matrix
			mov	t3, #8				' 8 of them
			mov	data, #$01			' bit mask for the address lines
:loop			rdbyte	kbd_data, t2			' next keyboard row entry
			add	t2, #1
			test	port, data		wz	' test port address A0, A1, A2 ...
		if_nz	or	t1, kbd_data			' if address bit was set, OR the data
			rol	data, #1			' rotate mask left
			djnz	t3, #:loop			' 8 times
			or	iocmd, t1			' return it in the command word bits 7..0
			jmp	#command_0

			fit	$1f0

io_command		long	0
portff			long	%00011000
irq			long	0
sync			long	0
colourram		byte	15[1024]
fontram			byte	0[1024]
videoram		byte	0[40*108]
startaddr		long	$4400	' video base address

psg			byte
			byte	0	' channel A fine tune
			byte	0	' channel A coarse tune
			byte	0	' channel B fine tune
			byte	0	' channel B coarse tune
			byte	0	' channel C fine tune
			byte	0	' channel C coarse tune
			byte	0	' noise period
			byte	$3f	' enable register
			byte	0	' channel A volume
			byte	0	' channel B volume
			byte	0	' channel C volume
			byte	0	' envelope fine tune
			byte	0	' envelope coarse tune
			byte	0	' envelope shape
			byte	0	' port A value
			byte	0	' port B value

crtc			byte
			byte	53	' 00 - horizontal total characters
			byte	40	' 01 - horizontal displayed characters per line
			byte	44	' 02 - horizontal synch position
			byte	1	' 03 - horizontal synch width in characters
			byte	27	' 04 - vertical total lines
			byte	0	' 05 - vertical total adjust (scan lines)
			byte	25	' 06 - vertical displayed rows
			byte	0	' 07 - vertical synch position (character rows)
			byte	0	' 08 - interlace mode
			byte	7	' 09 - maximum scan line address
			byte	0	' 0a - cursor start (scan line)
			byte	7	' 0b - cursor end (scan line)
			byte	4	' 0c - start address (MSB)
			byte	0	' 0d - start address (LSB)
			byte	4	' 0e - cursor address (MSB) (RW)
			byte	0	' 0f - cursor address (LSB) (RW)
			byte	0	' 10 - light pen (MSB) (RO)
			byte	0	' 11 - light pen (LSB) (RO)

msg_break		byte	"BC:",0
			byte	" DE:",0
			byte	" HL:",0
			byte	" AF:",0
			byte	" BC'",0
			byte	" DE'",0
			byte	" HL'",0
			byte	" AF'",0
			byte	13,10, "R:",0
			byte	" IFF1:",0
			byte	" IFF2:",0
			byte	" IM:",0
			byte	" I:",0
			byte	" IX:",0
			byte	" IY:",0
			byte	" SP:",0
			byte	" PC:",0
			byte	" = ",0

''      MIT LICENSE
{{
'  Permission is hereby granted, free of charge, to any person obtaining
'  a copy of this software and associated documentation files
'  (the "Software"), to deal in the Software without restriction,
'  including without limitation the rights to use, copy, modify, merge,
'  publish, distribute, sublicense, and/or sell copies of the Software,
'  and to permit persons to whom the Software is furnished to do so,
'  subject to the following conditions:
'
'  The above copyright notice and this permission notice shall be included
'  in all copies or substantial portions of the Software.
'
'  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
'  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
'  MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
'  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
'  CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
'  TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
'  SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
}}
